﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CSPspEmu.Core.Cpu;
using CSPspEmu.Core.Cpu.Emiter;
using CSPspEmu.Hle.Managers;
using CSPspEmu.Core;
using System.Reflection;
using CSPspEmu.Core.Gpu;
using CSPspEmu.Core.Rtc;
using CSPspEmu.Core.Display;
using CSPspEmu.Core.Audio;
using CSPspEmu.Core.Controller;
using CSPspEmu.Core.Crypto;
using CSPspEmu.Core.Battery;
using CSPspEmu.Hle.Vfs.MemoryStick;
using CSPspEmu.Hle.Vfs.Local;
using CSPspEmu.Hle.Vfs;

namespace CSPspEmu.Hle
{
	public class HleState : PspEmulatorComponent
	{
		public bool IsRunning;
		public CpuProcessor CpuProcessor;
		public GpuProcessor GpuProcessor;
		public PspRtc PspRtc;
		public PspDisplay PspDisplay;
		public PspAudio PspAudio;
		public PspController PspController;
		public PspConfig PspConfig;
		public Battery PspBattery;
		
		public MipsEmiter MipsEmiter;

		// Hle Managers
		public HleThreadManager ThreadManager;
		public HleSemaphoreManager SemaphoreManager;
		public HleEventFlagManager EventFlagManager;
		public HleMemoryManager MemoryManager;
		public HleModuleManager ModuleManager;
		public HleCallbackManager CallbackManager;
		public HleIoManager HleIoManager;
		public HleRegistryManager HleRegistryManager;
		public HleOutputHandler HleOutputHandler;
		public HleInterruptManager HleInterruptManager;
		public HleInterop HleInterop;
		public Kirk Kirk;

		/*
		public string MemoryStickRootLocalFolder
		{
			get
			{
				//var Mountable = (HleIoDriverMountable)HleIoManager.GetDriver("ms:");
				//var MemoryStick = (HleIoDriverMemoryStick)Mountable.GetMount("/");

				var MemoryStick = (HleIoDriverMemoryStick)HleIoManager.GetDriver("ms:");
				var Mountable = (HleIoDriverMountable)MemoryStick.ParentDriver;
				var Local = (HleIoDriverLocalFileSystem)(Mountable.GetMount("/"));
				return Local.BasePath;
			}
		}
		*/

		public override void InitializeComponent()
		{
			this.IsRunning = true;
			this.CpuProcessor = PspEmulatorContext.GetInstance<CpuProcessor>();
			this.GpuProcessor = PspEmulatorContext.GetInstance<GpuProcessor>();
			this.PspAudio = PspEmulatorContext.GetInstance<PspAudio>();
			this.PspConfig = PspEmulatorContext.PspConfig;
			this.PspRtc = PspEmulatorContext.GetInstance<PspRtc>();
			this.PspDisplay = PspEmulatorContext.GetInstance<PspDisplay>();
			this.PspController = PspEmulatorContext.GetInstance<PspController>();
			this.PspBattery = PspEmulatorContext.GetInstance<Battery>();

			this.MipsEmiter = new MipsEmiter();
			this.Kirk = new Kirk();
			this.Kirk.kirk_init();

			this.HleOutputHandler = PspEmulatorContext.GetInstance<HleOutputHandler>();

			// @TODO FIX! New Instances!?
			this.ThreadManager = PspEmulatorContext.GetInstance<HleThreadManager>();
			this.SemaphoreManager = PspEmulatorContext.GetInstance<HleSemaphoreManager>();
			this.EventFlagManager = PspEmulatorContext.GetInstance<HleEventFlagManager>();
			this.MemoryManager = new HleMemoryManager(this.CpuProcessor.Memory);
			this.ModuleManager = PspEmulatorContext.GetInstance<HleModuleManager>();
			this.CallbackManager = PspEmulatorContext.GetInstance<HleCallbackManager>();
			this.HleIoManager = PspEmulatorContext.GetInstance<HleIoManager>();
			this.HleRegistryManager = PspEmulatorContext.GetInstance<HleRegistryManager>();
			this.HleInterruptManager = PspEmulatorContext.GetInstance<HleInterruptManager>();
			this.HleInterop = PspEmulatorContext.GetInstance<HleInterop>();
		}
	}
}
